@@ -12,7 +12,7 @@
 #       as published by the Free Software Foundation.
 #
 #       This program is distributed in the hope that it will be useful,
-#       but WITHOUT ANY WARRANTY; without even the implied warranty of
+#       but WITHOUT ANY WARRANTY; without even the implied warranty ofs
 #       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 #       GNU General Public License for more details.
 #
@@ -30,7 +30,7 @@ use Shorewall::Config qw(:DEFAULT :internal);
 use Shorewall::IPAddrs;
 use Shorewall::Zones;
 use Shorewall::Chains qw(:DEFAULT :internal);
-use Shorewall::Policy;
+use Shorewall::Rules;
 use Shorewall::Proc;
 
 use strict;
@@ -45,7 +45,7 @@ our @EXPORT = qw( process_tos
 		  generate_matrix
 		  );
 our @EXPORT_OK = qw( initialize );
-our $VERSION = '4.4_16';
+our $VERSION = '4.4_19';
 
 our $family;
 
@@ -1036,13 +1036,40 @@ sub add_interface_jumps {
 	my $outputref    = $filter_table->{output_chain $interface};
 	my $interfaceref = find_interface($interface);
 
-	add_rule ( $filter_table->{FORWARD}, match_source_dev( $interface) . match_dest_dev( $interface) . '-j ACCEPT' ) unless $interfaceref->{nets} || ! $interfaceref->{options}{bridge};
+	if ( $interfaceref->{options}{port} ) {
+	    my $bridge = $interfaceref->{bridge};
+	    add_rule ( $filter_table->{forward_chain $bridge},
+		       match_source_dev( $interface, 1) . match_dest_dev( $interface, 1) . '-j ACCEPT'
+		     ) unless $interfaceref->{nets} || ! $interfaceref->{options}{bridge};
 
-	add_jump( $filter_table->{FORWARD} , $forwardref , 0, match_source_dev( $interface ) ) unless $forward_jump_added{$interface} || ! use_forward_chain $interface, $forwardref;
-	add_jump( $filter_table->{INPUT}   , $inputref ,   0, match_source_dev( $interface ) ) unless $input_jump_added{$interface}   || ! use_input_chain $interface, $inputref;
+	    add_jump( $filter_table->{forward_chain $bridge} ,
+		      $forwardref , 
+		      0, 
+		      match_source_dev( $interface, 1 )
+		    ) unless $forward_jump_added{$interface} || ! use_forward_chain $interface, $forwardref;
 
-	unless ( $output_jump_added{$interface} || ! use_output_chain $interface, $outputref ) {
-	    add_jump $filter_table->{OUTPUT} , $outputref , 0, match_dest_dev( $interface ) unless get_interface_option( $interface, 'port' );
+	    add_jump( $filter_table->{input_chain $bridge },
+		      $inputref ,
+		      0,
+		      match_source_dev( $interface, 1 )
+		    ) unless $input_jump_added{$interface}   || ! use_input_chain $interface, $inputref;
+
+	    unless ( $output_jump_added{$interface} || ! use_output_chain $interface, $outputref ) {
+		add_jump( $filter_table->{output_chain $bridge} ,
+			  $outputref ,
+			  0 ,
+			  match_dest_dev( $interface, 1 ) )
+		    unless get_interface_option( $interface, 'port' );
+	    }
+	} else {
+	    add_rule ( $filter_table->{FORWAR}, match_source_dev( $interface) . match_dest_dev( $interface) . '-j ACCEPT' ) unless $interfaceref->{nets} || ! $interfaceref->{options}{bridge};
+
+	    add_jump( $filter_table->{FORWARD} , $forwardref , 0, match_source_dev( $interface ) ) unless $forward_jump_added{$interface} || ! use_forward_chain $interface, $forwardref;
+	    add_jump( $filter_table->{INPUT}   , $inputref ,   0, match_source_dev( $interface ) ) unless $input_jump_added{$interface}   || ! use_input_chain $interface, $inputref;
+
+	    unless ( $output_jump_added{$interface} || ! use_output_chain $interface, $outputref ) {
+		add_jump $filter_table->{OUTPUT} , $outputref , 0, match_dest_dev( $interface ) unless get_interface_option( $interface, 'port' );
+	    }
 	}
     }
 
@@ -1077,6 +1104,7 @@ sub generate_matrix() {
     our %input_jump_added   = ();
     our %output_jump_added  = ();
     our %forward_jump_added = ();
+    my  %ipsec_jump_added   = ();
 
     progress_message2 'Generating Rule Matrix...';
     progress_message  '  Handling blacklisting and complex zones...';
@@ -1091,7 +1119,7 @@ sub generate_matrix() {
 	#
 	if ( $zoneref->{options}{in}{blacklist} ) {
 	    my $blackref = $filter_table->{blacklst};
-	    add_jump ensure_filter_chain( rules_chain( $zone, $_ ), 1 ) , $blackref , 0, $state, 0, -1 for firewall_zone, @vservers;
+	    add_jump ensure_rules_chain( rules_chain( $zone, $_ ) ) , $blackref , 0, $state, 0, -1 for firewall_zone, @vservers;
 
 	    if ( $simple ) {
 		#
@@ -1102,7 +1130,7 @@ sub generate_matrix() {
 		    my $ruleschainref = $filter_table->{$ruleschain};
 
 		    if ( ( $zone ne $zone1 || $ruleschainref->{referenced} ) && $ruleschainref->{policy} ne 'NONE' ) {
-			add_jump( ensure_filter_chain( $ruleschain, 1 ), $blackref, 0, $state, 0, -1 );
+			add_jump( ensure_rules_chain( $ruleschain ), $blackref, 0, $state, 0, -1 );
 		    }
 		}
 	    }
@@ -1110,14 +1138,14 @@ sub generate_matrix() {
 
 	if ( $zoneref->{options}{out}{blacklist} ) {
 	    my $blackref = $filter_table->{blackout};
-	    add_jump ensure_filter_chain( rules_chain( firewall_zone, $zone ), 1 ) , $blackref , 0, $state, 0, -1;
+	    add_jump ensure_rules_chain( rules_chain( firewall_zone, $zone ) ) , $blackref , 0, $state, 0, -1;
 
 	    for my $zone1 ( @zones, @vservers ) {
 		my $ruleschain    = rules_chain( $zone1, $zone );
 		my $ruleschainref = $filter_table->{$ruleschain};
 
 		if ( ( $zone ne $zone1 || $ruleschainref->{referenced} ) && $ruleschainref->{policy} ne 'NONE' ) {
-		    add_jump( ensure_filter_chain( $ruleschain, 1 ), $blackref, 0, $state, 0, -1 );
+		    add_jump( ensure_rules_chain( $ruleschain ), $blackref, 0, $state, 0, -1 );
 		}
 	    }
 	}
@@ -1143,12 +1171,31 @@ sub generate_matrix() {
 	    for my $interface ( sort { interface_number( $a ) <=> interface_number( $b ) } keys %$source_ref ) {
 		my $sourcechainref = $filter_table->{forward_chain $interface};
 		my $interfacematch = '';
+		my $interfaceref = find_interface $interface;
 
 		if ( use_forward_chain( $interface, $sourcechainref ) ) {
-		    add_jump $filter_table->{FORWARD} , $sourcechainref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+		    if ( $interfaceref->{ports} && $interfaceref->{options}{bridge} ) {
+			$interfacematch = match_source_dev $interface;
+			copy_rules( $sourcechainref, $frwd_ref, 1 ) unless $ipsec_jump_added{$zone}++;
+			$sourcechainref = $filter_table->{FORWARD};
+		    } elsif ( $interfaceref->{options}{port} ) {
+			add_jump( $filter_table->{ forward_chain $interfaceref->{bridge} } ,
+				  $sourcechainref ,
+				  0 ,
+				  match_source_dev( $interface , 1 ) )
+			    unless $forward_jump_added{$interface}++;
+		    } else {
+			add_jump $filter_table->{FORWARD} , $sourcechainref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+		    }
 		} else {
-		    $sourcechainref = $filter_table->{FORWARD};
-		    $interfacematch = match_source_dev $interface;
+		    if ( $interfaceref->{options}{port} ) {
+			$sourcechainref = $filter_table->{ forward_chain $interfaceref->{bridge} };
+			$interfacematch = match_source_dev $interface, 1;
+		    } else {
+			$sourcechainref = $filter_table->{FORWARD};
+			$interfacematch = match_source_dev $interface;
+		    }
+
 		    move_rules( $filter_table->{forward_chain $interface} , $frwd_ref );
 		}
 
@@ -1235,6 +1282,9 @@ sub generate_matrix() {
 	for my $typeref ( values %$source_hosts_ref ) {
 	    for my $interface ( sort { interface_number( $a ) <=> interface_number( $b ) } keys %$typeref ) {
 		my $arrayref = $typeref->{$interface};
+		my $interfaceref = find_interface $interface;
+		my $isport = $interfaceref->{options}{port};
+		my $bridge = $interfaceref->{bridge};
 
 		if ( get_physical( $interface ) eq '+' ) {
 		    #
@@ -1261,7 +1311,17 @@ sub generate_matrix() {
 
 			    if ( @vservers || use_output_chain( $interface, $interfacechainref ) || ( @{$interfacechainref->{rules}} && ! $chain1ref ) ) {
 				$outputref = $interfacechainref;
-				add_jump $filter_table->{OUTPUT}, $outputref, 0, match_dest_dev( $interface ) unless $output_jump_added{$interface}++;
+
+				if ( $isport ) {
+				    add_jump( $filter_table->{ output_chain $bridge },
+					      $outputref ,
+					      0 ,
+					      match_dest_dev( $interface, 1 ) )
+					unless $output_jump_added{$interface}++;
+				} else {
+				    add_jump $filter_table->{OUTPUT}, $outputref, 0, match_dest_dev( $interface ) unless $output_jump_added{$interface}++;
+				}
+
 				$use_output = 1;
 
 				unless ( lc $net eq IPv6_LINKLOCAL ) {
@@ -1269,6 +1329,9 @@ sub generate_matrix() {
 					generate_source_rules ( $outputref, $vzone, $zone, $dest );
 				    }
 				}
+			    } elsif ( $isport ) {
+				$outputref = $filter_table->{ output_chain $bridge };
+				$interfacematch = match_dest_dev $interface, 1;
 			    } else {
 				$outputref = $filter_table->{OUTPUT};
 				$interfacematch = match_dest_dev $interface;
@@ -1293,7 +1356,18 @@ sub generate_matrix() {
 			    # There are DNAT/REDIRECT rules with this zone as the source.
 			    # Add a jump from this source network to this zone's DNAT/REDIRECT chain
 			    #
-			    add_jump $preroutingref, source_exclusion( $exclusions, $dnatref), 0, join( '', match_source_dev( $interface), $source, $ipsec_in_match );
+			    add_jump( $preroutingref,
+				      source_exclusion( $exclusions, $dnatref),
+				      0,
+				      join( '', match_source_dev( $interface), $source, $ipsec_in_match ) );
+				      
+			    if ( get_physical( $interface ) eq '+' ) {
+				#
+				# The jump from the PREROUTING chain to dnat may not have been added above
+				# 
+				addnatjump 'PREROUTING', 'dnat', '' unless $preroutingref->{references}{PREROUTING};
+			    }
+				
 			    check_optimization( $dnatref ) if $source;
 			}
 
@@ -1323,7 +1397,17 @@ sub generate_matrix() {
 
 			if ( @vservers || use_input_chain( $interface, $interfacechainref ) || ! $chain2 || ( @{$interfacechainref->{rules}} && ! $chain2ref ) ) {
 			    $inputchainref = $interfacechainref;
-			    add_jump $filter_table->{INPUT}, $inputchainref, 0, match_source_dev($interface) unless $input_jump_added{$interface}++;
+			    
+			    if ( $isport ) {
+				add_jump( $filter_table->{ input_chain $bridge },
+					  $inputchainref ,
+					  0 ,
+					  match_source_dev($interface, 1) )
+				    unless $input_jump_added{$interface}++;
+			    } else {
+				add_jump $filter_table->{INPUT}, $inputchainref, 0, match_source_dev($interface) unless $input_jump_added{$interface}++;
+			    }
+
 			    $use_input = 1;
 
 			    unless ( lc $net eq IPv6_LINKLOCAL ) {
@@ -1332,6 +1416,9 @@ sub generate_matrix() {
 				    generate_dest_rules( $inputchainref, $target, $vzone, $source . $ipsec_in_match ) if $target;
 				}
 			    }
+			} elsif ( $isport ) {
+			    $inputchainref = $filter_table->{ input_chain $bridge };
+			    $interfacematch = match_source_dev $interface, 1;
 			} else {
 			    $inputchainref = $filter_table->{INPUT};
 			    $interfacematch = match_source_dev $interface;
@@ -1345,11 +1432,29 @@ sub generate_matrix() {
 			if ( $frwd_ref && $hostref->{ipsec} ne 'ipsec' ) {
 			    my $ref = source_exclusion( $exclusions, $frwd_ref );
 			    my $forwardref = $filter_table->{forward_chain $interface};
+
 			    if ( use_forward_chain $interface, $forwardref ) {
 				add_jump $forwardref , $ref, 0, join( '', $source, $ipsec_in_match );
-				add_jump $filter_table->{FORWARD} , $forwardref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+				
+				if ( $isport ) {
+				    add_jump( $filter_table->{ forward_chain $bridge } ,
+					      $forwardref ,
+					      0 ,
+					      match_source_dev( $interface , 1 ) )
+					unless $forward_jump_added{$interface}++;
+				} else {
+				    add_jump $filter_table->{FORWARD} , $forwardref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+				}
 			    } else {
-				add_jump $filter_table->{FORWARD} , $ref, 0, join( '', match_source_dev( $interface ) , $source, $ipsec_in_match );
+				if ( $isport ) {
+				    add_jump( $filter_table->{ forward_chain $bridge } ,
+					      $ref ,
+					      0 ,
+					      join( '', match_source_dev( $interface, 1 ) , $source, $ipsec_in_match ) );
+				} else {
+				    add_jump $filter_table->{FORWARD} , $ref, 0, join( '', match_source_dev( $interface ) , $source, $ipsec_in_match );
+				}
+
 				move_rules ( $forwardref , $frwd_ref );
 			    }
 			}
@@ -1461,6 +1566,7 @@ sub generate_matrix() {
 		#
 		for my $typeref ( values %$source_hosts_ref ) {
 		    for my $interface ( keys %$typeref ) {
+			my $interfaceref = find_interface $interface;
 			my $chain3ref;
 			my $match_source_dev = '';
 			my $forwardchainref = $filter_table->{forward_chain $interface};
@@ -1470,13 +1576,28 @@ sub generate_matrix() {
 			    # Either we must use the interface's forwarding chain or that chain has rules and we have nowhere to move them
 			    #
 			    $chain3ref = $forwardchainref;
-			    add_jump $filter_table->{FORWARD} , $chain3ref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+			    
+			    if ( $interfaceref->{options}{port} ) {
+				add_jump( $filter_table->{ forward_chain $interfaceref->{bridge} } ,
+					  $chain3ref,
+					  0 ,
+					  match_source_dev( $interface , 1 ) )
+				    unless $forward_jump_added{$interface}++;
+			    } else {
+				add_jump $filter_table->{FORWARD} , $chain3ref, 0 , match_source_dev( $interface ) unless $forward_jump_added{$interface}++;
+			    }
 			} else {
 			    #
 			    # Don't use the interface's forward chain -- move any rules in that chain to this rules chain
 			    #
-			    $chain3ref  = $filter_table->{FORWARD};
-			    $match_source_dev = match_source_dev $interface;
+			    if ( $interfaceref->{options}{port} ) {
+				$chain3ref  = $filter_table->{ forward_chain $interfaceref->{bridge} };
+				$match_source_dev = match_source_dev $interface, 1;
+			    } else {
+				$chain3ref  = $filter_table->{FORWARD};
+				$match_source_dev = match_source_dev $interface;
+			    }
+
 			    move_rules $forwardchainref, $chainref;
 			}
 
@@ -1766,7 +1887,7 @@ EOF
     emit 'delete_tc1' if $config{CLEAR_TC};
 
     emit( 'undo_routing',
-	  'restore_default_route'
+	  "restore_default_route $config{USE_DEFAULT_RT}"
 	  );
 
     my @chains = $config{ADMINISABSENTMINDED} ? qw/INPUT FORWARD/ : qw/INPUT OUTPUT FORWARD/;
@@ -1786,7 +1907,6 @@ EOF
 
     process_routestopped;
 
-    add_rule $input, '-i lo -j ACCEPT';
     add_rule $input, '-i lo -j ACCEPT';
 
     add_rule $output, '-o lo -j ACCEPT' unless $config{ADMINISABSENTMINDED};
